@@ -27,39 +27,19 @@
 
 from __future__ import absolute_import
 
-from ..util import APPLICATION
+from ..descriptor import open_device, FailedOpeningDeviceException
+from ..fido import FipsU2fController
+from ..oath import OathController
+from ..otp import OtpController
+from ..util import APPLICATION, TRANSPORT
 import click
+import logging
 
 
-@click.command()
-@click.pass_context
-def info(ctx):
-    """
-    Show general information.
+logger = logging.getLogger(__name__)
 
-    Displays information about the attached YubiKey such as serial number,
-    firmware version, applications, etc.
-    """
-    dev = ctx.obj['dev']
-    click.echo('Device type: {}'.format(dev.device_name))
-    click.echo('Serial number: {}'.format(
-        dev.serial or 'Not set or unreadable'))
-    if dev.version:
-        f_version = '.'.join(str(x) for x in dev.version)
-        click.echo('Firmware version: {}'.format(f_version))
-    else:
-        click.echo('Firmware version: Uncertain, re-run with only one '
-                   'YubiKey connected')
-    config = dev.config
-    if config.form_factor:
-        click.echo('Form factor: {!s}'.format(config.form_factor))
-    click.echo('Enabled USB interfaces: {}'.format(dev.mode))
-    if config.nfc_supported:
-        f_nfc = 'enabled' if config.nfc_enabled else 'disabled'
-        click.echo('NFC interface is {}.'.format(f_nfc))
-    if config.configuration_locked:
-        click.echo('Configured applications are protected by a lock code.')
-    click.echo()
+
+def print_app_status_table(config):
     rows = []
     for app in APPLICATION:
         if app & config.usb_supported:
@@ -106,3 +86,86 @@ def info(ctx):
     else:
         click.echo('{}'.format(f_apps))
     click.echo(f_table, nl=False)
+
+
+def get_fips_status_over_transport(serial, transport, controller_constructor):
+    try:
+        with open_device(transports=transport, serial=serial) as dev:
+            return controller_constructor(dev._driver).is_in_fips_mode
+    except FailedOpeningDeviceException as e:
+        logger.debug('Failed to open device', exc_info=e)
+        return False
+
+
+def get_overall_fips_status(serial, config):
+    statuses = {}
+
+    if config.usb_enabled & APPLICATION.OTP:
+        statuses['OTP'] = get_fips_status_over_transport(
+            serial, TRANSPORT.OTP, OtpController)
+    else:
+        statuses['OTP'] = False
+
+    if config.usb_enabled & APPLICATION.OATH:
+        statuses['OATH'] = get_fips_status_over_transport(
+            serial, TRANSPORT.CCID, OathController)
+    else:
+        statuses['OATH'] = False
+
+    if config.usb_enabled & APPLICATION.U2F:
+        statuses['FIDO U2F'] = get_fips_status_over_transport(
+            serial, TRANSPORT.FIDO, FipsU2fController)
+    else:
+        statuses['FIDO U2F'] = False
+
+    return statuses
+
+
+@click.command()
+@click.pass_context
+def info(ctx):
+    """
+    Show general information.
+
+    Displays information about the attached YubiKey such as serial number,
+    firmware version, applications, etc.
+    """
+    dev = ctx.obj['dev']
+
+    if dev.is_fips:
+        fips_status = get_overall_fips_status(dev.serial, dev.config)
+
+    click.echo('Device type: {}'.format(dev.device_name))
+    click.echo('Serial number: {}'.format(
+        dev.serial or 'Not set or unreadable'))
+    if dev.version:
+        f_version = '.'.join(str(x) for x in dev.version)
+        click.echo('Firmware version: {}'.format(f_version))
+    else:
+        click.echo('Firmware version: Uncertain, re-run with only one '
+                   'YubiKey connected')
+
+    config = dev.config
+    if config.form_factor:
+        click.echo('Form factor: {!s}'.format(config.form_factor))
+    click.echo('Enabled USB interfaces: {}'.format(dev.mode))
+    if config.nfc_supported:
+        f_nfc = 'enabled' if config.nfc_enabled else 'disabled'
+        click.echo('NFC interface is {}.'.format(f_nfc))
+    if config.configuration_locked:
+        click.echo('Configured applications are protected by a lock code.')
+    click.echo()
+
+    print_app_status_table(config)
+
+    if dev.is_fips:
+        click.echo()
+
+        click.echo('FIPS Approved Mode: {}'.format(
+            'Yes' if all(fips_status.values()) else 'No'))
+
+        status_keys = list(fips_status.keys())
+        status_keys.sort()
+        for status_key in status_keys:
+            click.echo('  {}: {}'.format(
+                status_key, 'Yes' if fips_status[status_key] else 'No'))
