@@ -31,8 +31,8 @@ import logging
 import click
 from ..util import TRANSPORT
 from ..opgp import OpgpController, KEY_SLOT, TOUCH_MODE
-from ..driver_ccid import APDUError, SW_APPLICATION_NOT_FOUND
-from .util import click_force_option, click_skip_on_help
+from ..driver_ccid import APDUError, SW
+from .util import click_force_option, click_postpone_execution
 
 
 logger = logging.getLogger(__name__)
@@ -79,16 +79,25 @@ def int_in_range(minval, maxval):
 
 @click.group()
 @click.pass_context
-@click_skip_on_help
+@click_postpone_execution
 def openpgp(ctx):
     """
     Manage OpenPGP Application.
+
+    Examples:
+
+    \b
+      Set the retries for PIN, Reset Code and Admin PIN to 10:
+      $ ykman openpgp set-retries 10 10 10
+
+    \b
+      Require touch to use the authentication key:
+      $ ykman openpgp touch aut on
     """
     try:
-        controller = OpgpController(ctx.obj['dev'].driver)
-        ctx.obj['controller'] = controller
+        ctx.obj['controller'] = OpgpController(ctx.obj['dev'].driver)
     except APDUError as e:
-        if e.sw == SW_APPLICATION_NOT_FOUND:
+        if e.sw == SW.NOT_FOUND:
             ctx.fail("The OpenPGP application can't be found on this "
                      'YubiKey.')
         logger.debug('Failed to load OpenPGP Application', exc_info=e)
@@ -161,9 +170,9 @@ def touch(ctx, key, policy, admin_pin, force):
         ctx.fail('A FIXED policy cannot be changed!')
 
     force or click.confirm('Set touch policy of {.name} key to {.name}?'.format(
-        key, policy), abort=True)
+        key, policy), abort=True, err=True)
     if admin_pin is None:
-        admin_pin = click.prompt('Enter admin PIN', hide_input=True)
+        admin_pin = click.prompt('Enter admin PIN', hide_input=True, err=True)
     controller.set_touch(key, policy, admin_pin.encode('utf8'))
     click.echo('Touch policy successfully set.')
 
@@ -189,7 +198,7 @@ def set_pin_retries(ctx, pw_attempts, admin_pin, force):
         click.echo('WARNING: Setting PIN retries will reset the values for all '
                    '3 PINs!')
     force or click.confirm('Set PIN retry counters to: {} {} {}?'.format(
-        *pw_attempts), abort=True)
+        *pw_attempts), abort=True, err=True)
     controller.set_pin_retries(*(pw_attempts + (admin_pin.encode('utf8'),)))
     click.echo('PIN retries successfully set.')
     if resets_pins:
