@@ -29,7 +29,7 @@ from __future__ import absolute_import
 
 from .util import (
     click_force_option, click_callback, click_parse_b32_key,
-    click_skip_on_help, prompt_for_touch, UpperCaseChoice)
+    click_postpone_execution, prompt_for_touch, UpperCaseChoice)
 from ..util import (
     TRANSPORT, generate_static_pw, modhex_decode,
     modhex_encode, parse_key, parse_b32_key)
@@ -82,16 +82,16 @@ def _confirm_slot_overwrite(controller, slot):
     if slot == 1 and slot1:
         click.confirm(
             'Slot 1 is already configured. Overwrite configuration?',
-            abort=True)
+            abort=True, err=True)
     if slot == 2 and slot2:
         click.confirm(
             'Slot 2 is already configured. Overwrite configuration?',
-            abort=True)
+            abort=True, err=True)
 
 
 @click.group()
 @click.pass_context
-@click_skip_on_help
+@click_postpone_execution
 @click.option(
     '--access-code', required=False, metavar='HEX',
     help='A 6 byte access code. Set to empty to use a prompt for input.')
@@ -106,12 +106,31 @@ def otp(ctx, access_code):
     prevents the configuration to be overwritten without the access code
     provided. Mode switching the YubiKey is not possible when a slot is
     configured with an access code.
+
+    Examples:
+
+    \b
+      Swap the configurations between the two slots:
+      $ ykman otp swap
+
+    \b
+      Program a random challenge-response credential to slot 2:
+      $ ykman otp chalresp --generate 2
+
+    \b
+      Program a Yubico OTP credential to slot 2, using the serial as public id:
+      $ ykman otp yubiotp 1 --serial-public-id
+
+    \b
+      Program a random 38 characters long static password to slot 2:
+      $ ykman otp static --generate 2 --length 38
     """
 
     ctx.obj['controller'] = OtpController(ctx.obj['dev'].driver)
     if access_code is not None:
         if access_code == '':
-            access_code = click.prompt('Enter access code', show_default=False)
+            access_code = click.prompt(
+                'Enter access code', show_default=False, err=True)
 
         try:
             access_code = parse_access_code_hex(access_code)
@@ -196,7 +215,7 @@ def delete(ctx, slot, force):
         ctx.fail('Not possible to delete an empty slot.')
     force or click.confirm(
         'Do you really want to delete'
-        ' the configuration of slot {}?'.format(slot), abort=True)
+        ' the configuration of slot {}?'.format(slot), abort=True, err=True)
     click.echo('Deleting the configuration of slot {}...'.format(slot))
     try:
         controller.zap_slot(slot)
@@ -260,7 +279,7 @@ def yubiotp(ctx, slot, public_id, private_id, key, no_enter, force,
                 'Public ID not given. Please remove the --force flag, or '
                 'add the --serial-public-id flag or --public-id option.')
         else:
-            public_id = click.prompt('Enter public ID')
+            public_id = click.prompt('Enter public ID', err=True)
 
     try:
         public_id = modhex_decode(public_id)
@@ -278,7 +297,7 @@ def yubiotp(ctx, slot, public_id, private_id, key, no_enter, force,
                 'Private ID not given. Please remove the --force flag, or '
                 'add the --generate-private-id flag or --private-id option.')
         else:
-            private_id = click.prompt('Enter private ID')
+            private_id = click.prompt('Enter private ID', err=True)
             private_id = a2b_hex(private_id)
 
     if not key:
@@ -291,11 +310,11 @@ def yubiotp(ctx, slot, public_id, private_id, key, no_enter, force,
             ctx.fail('Secret key not given. Please remove the --force flag, or '
                      'add the --generate-key flag or --key option.')
         else:
-            key = click.prompt('Enter secret key')
+            key = click.prompt('Enter secret key', err=True)
             key = a2b_hex(key)
 
     force or click.confirm('Program an OTP credential in slot {}?'.format(slot),
-                           abort=True)
+                           abort=True, err=True)
     try:
         controller.program_otp(slot, key, public_id, private_id, not no_enter)
     except YkpersError as e:
@@ -332,6 +351,7 @@ def static(
     preferred keyboard layout.
     """
 
+    controller = ctx.obj['controller']
     keyboard_layout = KEYBOARD_LAYOUT[keyboard_layout]
 
     if password and len(password) > 38:
@@ -340,12 +360,10 @@ def static(
         ctx.fail('Provide a length for the generated password.')
 
     if not password and not generate:
-        password = click.prompt('Enter a static password')
+        password = click.prompt('Enter a static password', err=True)
     elif not password and generate:
         password = generate_static_pw(length, keyboard_layout).decode()
 
-    controller = ctx.obj['controller']
-
     if not force:
         _confirm_slot_overwrite(controller, slot)
     try:
@@ -390,25 +408,24 @@ def chalresp(ctx, slot, key, totp, touch, force, generate):
                      'set the KEY argument or set the --generate flag.')
         elif totp:
             while True:
-                key = click.prompt('Enter a secret key (base32)')
+                key = click.prompt('Enter a secret key (base32)', err=True)
                 try:
                     key = parse_b32_key(key)
                     break
                 except Exception as e:
                     click.echo(e)
-                    pass
         else:
             if generate:
                 key = os.urandom(20)
                 click.echo('Using a randomly generated key: {}'.format(
                     b2a_hex(key).decode('ascii')))
             else:
-                key = click.prompt('Enter a secret key')
+                key = click.prompt('Enter a secret key', err=True)
                 key = parse_key(key)
 
     cred_type = 'TOTP' if totp else 'challenge-response'
     force or click.confirm('Program a {} credential in slot {}?'
-                           .format(cred_type, slot), abort=True)
+                           .format(cred_type, slot), abort=True, err=True)
     try:
         controller.program_chalresp(slot, key, touch)
     except YkpersError as e:
@@ -430,7 +447,7 @@ def calculate(ctx, slot, challenge, totp, digits):
     Perform a challenge-response operation.
 
     Send a challenge (in hex) to a YubiKey slot with a challenge-response
-credential, and read the response. Supports output as a OATH-TOTP code.
+    credential, and read the response. Supports output as a OATH-TOTP code.
     """
     controller = ctx.obj['controller']
     if not challenge and not totp:
@@ -490,16 +507,16 @@ def hotp(ctx, slot, key, digits, counter, no_enter, force):
     controller = ctx.obj['controller']
     if not key:
         while True:
-            key = click.prompt('Enter a secret key (base32)')
+            key = click.prompt('Enter a secret key (base32)', err=True)
             try:
                 key = parse_b32_key(key)
                 break
             except Exception as e:
                 click.echo(e)
-                pass
 
     force or click.confirm(
-        'Program a HOTP credential in slot {}?'.format(slot), abort=True)
+        'Program a HOTP credential in slot {}?'.format(slot), abort=True,
+        err=True)
     try:
         controller.program_hotp(
             slot, key, counter, int(digits) == 8, not no_enter)
@@ -544,7 +561,7 @@ def settings(ctx, slot, new_access_code, delete_access_code, enter, pacing,
     if new_access_code is not None:
         if new_access_code == '':
             new_access_code = click.prompt(
-                'Enter new access code', show_default=False)
+                'Enter new access code', show_default=False, err=True)
 
         try:
             new_access_code = parse_access_code_hex(new_access_code)
@@ -553,7 +570,8 @@ def settings(ctx, slot, new_access_code, delete_access_code, enter, pacing,
 
     force or click.confirm(
         'Update the settings for slot {}? '
-        'All existing settings will be overwritten.'.format(slot), abort=True)
+        'All existing settings will be overwritten.'.format(slot), abort=True,
+        err=True)
     click.echo('Updating settings for slot {}...'.format(slot))
 
     if pacing is not None:
