@@ -27,16 +27,19 @@
 
 from __future__ import absolute_import
 
-from ..util import TRANSPORT, parse_private_key, parse_certificate
+from ..util import (
+    TRANSPORT, get_leaf_certificates, parse_private_key, parse_certificates)
 from ..piv import (
-    PivController, ALGO, OBJ, SW, SLOT, PIN_POLICY, TOUCH_POLICY,
+    PivController, ALGO, OBJ, SLOT, PIN_POLICY, TOUCH_POLICY,
     DEFAULT_MANAGEMENT_KEY, generate_random_management_key)
-from ..driver_ccid import APDUError, SW_APPLICATION_NOT_FOUND
+from ..piv import (
+    AuthenticationBlocked, AuthenticationFailed, KeypairMismatch,
+    UnsupportedAlgorithm, WrongPin, WrongPuk)
+from ..driver_ccid import APDUError, SW
 from .util import (
-    click_force_option, click_skip_on_help, click_callback, prompt_for_touch,
-    UpperCaseChoice)
+    click_force_option, click_postpone_execution, click_callback,
+    prompt_for_touch, UpperCaseChoice)
 from cryptography import x509
-from cryptography.exceptions import InvalidSignature
 from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.backends import default_backend
 from binascii import b2a_hex, a2b_hex
@@ -71,11 +74,11 @@ def click_parse_management_key(ctx, param, val):
     try:
         key = a2b_hex(val)
         if key and len(key) != 24:
-            return ValueError('Management key must be exactly 24 bytes '
-                              '(48 hexadecimal digits) long.')
+            raise ValueError('Management key must be exactly 24 bytes '
+                             '(48 hexadecimal digits) long.')
         return key
     except Exception:
-        return ValueError(val)
+        raise ValueError(val)
 
 
 click_slot_argument = click.argument('slot', callback=click_parse_piv_slot)
@@ -101,16 +104,31 @@ click_touch_policy_option = click.option(
 
 @click.group()
 @click.pass_context
-@click_skip_on_help
+@click_postpone_execution
 def piv(ctx):
     """
     Manage PIV Application.
+
+    Examples:
+
+    \b
+      Generate an ECC P-256 private key and a self-signed certificate in
+      slot 9a:
+      $ ykman piv generate-key --algorithm ECCP256 9a pubkey.pem
+      $ ykman piv generate-certificate --subject "yubico" 9a pubkey.pem
+
+    \b
+      Change the PIN from 123456 to 654321:
+      $ ykman piv change-pin --pin 123456 --new-pin 654321
+
+    \b
+      Reset all PIV data and restore default settings:
+      $ ykman piv reset
     """
     try:
-        controller = PivController(ctx.obj['dev'].driver)
-        ctx.obj['controller'] = controller
+        ctx.obj['controller'] = PivController(ctx.obj['dev'].driver)
     except APDUError as e:
-        if e.sw == SW_APPLICATION_NOT_FOUND:
+        if e.sw == SW.NOT_FOUND:
             ctx.fail("The PIV application can't be found on this YubiKey.")
         raise
 
@@ -217,26 +235,30 @@ def generate_key(
     PUBLIC-KEY  File containing the generated public key. Use '-' to use stdout.
     """
 
+    dev = ctx.obj['dev']
     controller = ctx.obj['controller']
 
     _ensure_authenticated(ctx, controller, pin, management_key)
 
-    algorithm = ALGO.from_string(algorithm)
+    algorithm_id = ALGO.from_string(algorithm)
 
     if pin_policy:
         pin_policy = PIN_POLICY.from_string(pin_policy)
     if touch_policy:
         touch_policy = TOUCH_POLICY.from_string(touch_policy)
 
-    _check_algorithm(ctx, controller, algorithm)
-    _check_pin_policy(ctx, controller, pin_policy)
+    _check_pin_policy(ctx, dev, controller, pin_policy)
     _check_touch_policy(ctx, controller, touch_policy)
 
-    public_key = controller.generate_key(
-        slot,
-        algorithm,
-        pin_policy,
-        touch_policy)
+    try:
+        public_key = controller.generate_key(
+            slot,
+            algorithm_id,
+            pin_policy,
+            touch_policy)
+    except UnsupportedAlgorithm:
+        ctx.fail('Algorithm {} is not supported by this '
+                 'YubiKey.'.format(algorithm))
 
     key_encoding = format
     public_key_output.write(public_key.public_bytes(
@@ -251,9 +273,14 @@ def generate_key(
 @click_pin_option
 @click.option(
     '-p', '--password', help='A password may be needed to decrypt the data.')
+@click.option(
+    '--no-verify', 'verify', is_flag=True, default=False,
+    callback=lambda ctx, param, value: not value,
+    help='Skip verifying that the certificate matches the private key in the '
+         'slot.')
 @click.argument('cert', type=click.File('rb'), metavar='CERTIFICATE')
 def import_certificate(
-        ctx, slot, management_key, pin, cert, password):
+        ctx, slot, management_key, pin, cert, password, verify):
     """
     Import a X.509 certificate.
 
@@ -272,13 +299,14 @@ def import_certificate(
         if password is not None:
             password = password.encode()
         try:
-            cert = parse_certificate(data, password)
+            certs = parse_certificates(data, password)
         except (ValueError, TypeError):
             if password is None:
                 password = click.prompt(
                     'Enter password to decrypt certificate',
                     default='', hide_input=True,
-                    show_default=False)
+                    show_default=False,
+                    err=True)
                 continue
             else:
                 password = None
@@ -286,7 +314,30 @@ def import_certificate(
             continue
         break
 
-    controller.import_certificate(slot, cert)
+    if len(certs) > 1:
+        #  If multiple certs, only import leaf.
+        #  Leaf is the cert with a subject that is not an issuer in the chain.
+        leafs = get_leaf_certificates(certs)
+        cert_to_import = leafs[0]
+    else:
+        cert_to_import = certs[0]
+
+    def do_import(retry=True):
+        try:
+            controller.import_certificate(slot, cert_to_import, verify=verify)
+
+        except KeypairMismatch:
+            ctx.fail('This certificate is not tied to the private key in the '
+                     '{} slot.'.format(slot.name))
+
+        except APDUError as e:
+            if e.sw == SW.SECURITY_CONDITION_NOT_SATISFIED and retry:
+                _verify_pin(ctx, controller, pin)
+                do_import(retry=False)
+            else:
+                raise
+
+    do_import()
 
 
 @piv.command('import-key')
@@ -311,6 +362,7 @@ def import_key(
     SLOT        PIV slot to import the private key to.
     PRIVATE-KEY File containing the private key. Use '-' to use stdin.
     """
+    dev = ctx.obj['dev']
     controller = ctx.obj['controller']
     _ensure_authenticated(ctx, controller, pin, management_key)
 
@@ -326,7 +378,8 @@ def import_key(
                 password = click.prompt(
                     'Enter password to decrypt key',
                     default='', hide_input=True,
-                    show_default=False)
+                    show_default=False,
+                    err=True)
                 continue
             else:
                 password = None
@@ -339,9 +392,9 @@ def import_key(
     if touch_policy:
         touch_policy = TOUCH_POLICY.from_string(touch_policy)
 
-    _check_pin_policy(ctx, controller, pin_policy)
+    _check_pin_policy(ctx, dev, controller, pin_policy)
     _check_touch_policy(ctx, controller, touch_policy)
-    _check_key_size(ctx, private_key)
+    _check_key_size(ctx, controller, private_key)
 
     controller.import_key(
             slot,
@@ -449,7 +502,7 @@ def set_pin_retries(ctx, management_key, pin, pin_retries, puk_retries, force):
     click.echo('WARNING: This will reset the PIN and PUK to the factory '
                'defaults!')
     force or click.confirm('Set PIN and PUK retry counters to: {} {}?'.format(
-        pin_retries, puk_retries), abort=True)
+        pin_retries, puk_retries), abort=True, err=True)
     try:
         controller.set_pin_retries(pin_retries, puk_retries)
         click.echo('Default PINs are set.')
@@ -506,9 +559,6 @@ def generate_certificate(
                      exc_info=e)
         ctx.fail('Certificate generation failed.')
 
-    except InvalidSignature:
-        ctx.fail('Invalid signature, certificate not imported.')
-
 
 @piv.command('generate-csr')
 @click.pass_context
@@ -582,13 +632,19 @@ def change_pin(ctx, pin, new_pin):
     if not new_pin:
         new_pin = click.prompt(
             'Enter your new PIN', default='', hide_input=True,
-            show_default=False, confirmation_prompt=True)
+            show_default=False, confirmation_prompt=True, err=True)
     try:
         controller.change_pin(pin, new_pin)
-    except APDUError as e:
-        logger.error('Failed to change PIN', exc_info=e)
-        ctx.fail('Changing the PIN failed.')
-    click.echo('New PIN set.')
+        click.echo('New PIN set.')
+
+    except AuthenticationBlocked as e:
+        logger.debug('PIN is blocked.', exc_info=e)
+        ctx.fail('PIN is blocked.')
+
+    except WrongPin as e:
+        logger.debug(
+            'Failed to change PIN, %d tries left', e.tries_left, exc_info=e)
+        ctx.fail('PIN change failed - %d tries left.' % e.tries_left)
 
 
 @piv.command('change-puk')
@@ -607,15 +663,21 @@ def change_puk(ctx, puk, new_puk):
     if not new_puk:
         new_puk = click.prompt(
             'Enter your new PUK', default='', hide_input=True,
-            show_default=False, confirmation_prompt=True)
+            show_default=False, confirmation_prompt=True,
+            err=True)
 
-    (success, retries) = controller.change_puk(puk, new_puk)
-
-    if success:
+    try:
+        controller.change_puk(puk, new_puk)
         click.echo('New PUK set.')
-    else:
-        logger.debug('Failed to change PUK, %d tries left', retries)
-        ctx.fail('PUK change failed - %d tries left.' % retries)
+
+    except AuthenticationBlocked as e:
+        logger.debug('PUK is blocked.', exc_info=e)
+        ctx.fail('PUK is blocked.')
+
+    except WrongPuk as e:
+        logger.debug(
+            'Failed to change PUK, %d tries left', e.tries_left, exc_info=e)
+        ctx.fail('PUK change failed - %d tries left.' % e.tries_left)
 
 
 @piv.command('change-management-key')
@@ -651,7 +713,7 @@ def change_management_key(
     """
     controller = ctx.obj['controller']
 
-    _ensure_authenticated(
+    pin_verified = _ensure_authenticated(
         ctx, controller, pin, management_key,
         require_pin_and_key=protect,
         mgm_key_prompt='Enter your current management key '
@@ -667,14 +729,14 @@ def change_management_key(
         ctx.fail('Require touch not supported on this YubiKey.')
 
     # If an old stored key needs to be cleared, the PIN is needed.
-    if not protect and controller.has_stored_key:
+    if not pin_verified and controller.has_stored_key:
         if pin:
             _verify_pin(ctx, controller, pin, no_prompt=force)
-        else:
-            force or click.confirm(
+        elif not force:
+            click.confirm(
                     'The current management key is stored on the YubiKey'
                     ' and will not be cleared if no PIN is provided. Continue?',
-                    abort=True)
+                    abort=True, err=True)
 
     if not new_management_key and not protect:
         if generate:
@@ -693,7 +755,7 @@ def change_management_key(
         else:
             new_management_key = click.prompt(
                 'Enter your new management key',
-                hide_input=True, confirmation_prompt=True)
+                hide_input=True, confirmation_prompt=True, err=True)
 
     if new_management_key and type(new_management_key) is not bytes:
         try:
@@ -722,17 +784,19 @@ def unblock_pin(ctx, puk, new_pin):
     controller = ctx.obj['controller']
     if not puk:
         puk = click.prompt(
-            'Enter PUK', default='', show_default=False, hide_input=True)
+            'Enter PUK', default='', show_default=False,
+            hide_input=True, err=True)
     if not new_pin:
         new_pin = click.prompt(
-            'Enter a new PIN', default='', show_default=False, hide_input=True)
+            'Enter a new PIN', default='',
+            show_default=False, hide_input=True, err=True)
     controller.unblock_pin(puk, new_pin)
 
 
 def _prompt_management_key(
         ctx, prompt='Enter a management key [blank to use default key]'):
     management_key = click.prompt(
-        prompt, default='', hide_input=True, show_default=False)
+        prompt, default='', hide_input=True, show_default=False, err=True)
     if management_key == '':
         return DEFAULT_MANAGEMENT_KEY
     try:
@@ -743,7 +807,7 @@ def _prompt_management_key(
 
 def _prompt_pin(ctx, prompt='Enter PIN'):
     return click.prompt(
-        prompt, default='', hide_input=True, show_default=False)
+        prompt, default='', hide_input=True, show_default=False, err=True)
 
 
 def _ensure_authenticated(
@@ -752,17 +816,22 @@ def _ensure_authenticated(
         mgm_key_prompt=None,
         no_prompt=False):
 
+    pin_verified = False
+
     if controller.has_protected_key:
         if not management_key:
-            _verify_pin(ctx, controller, pin, no_prompt=no_prompt)
+            pin_verified = _verify_pin(
+                ctx, controller, pin, no_prompt=no_prompt)
         else:
             _authenticate(ctx, controller, management_key, mgm_key_prompt,
                           no_prompt=no_prompt)
     else:
         if require_pin_and_key:
-            _verify_pin(ctx, controller, pin, no_prompt=no_prompt)
+            pin_verified = _verify_pin(
+                ctx, controller, pin, no_prompt=no_prompt)
         _authenticate(ctx, controller, management_key, mgm_key_prompt,
                       no_prompt=no_prompt)
+    return pin_verified
 
 
 def _verify_pin(ctx, controller, pin, no_prompt=False):
@@ -774,7 +843,12 @@ def _verify_pin(ctx, controller, pin, no_prompt=False):
 
     try:
         controller.verify(pin, touch_callback=prompt_for_touch)
-    except APDUError:
+        return True
+    except WrongPin as e:
+        ctx.fail('PIN verification failed, {} tries left.'.format(e.tries_left))
+    except AuthenticationBlocked as e:
+        ctx.fail('PIN is blocked.')
+    except Exception:
         ctx.fail('PIN verification failed.')
 
 
@@ -790,27 +864,23 @@ def _authenticate(ctx, controller, management_key, mgm_key_prompt,
                 management_key = _prompt_management_key(ctx, mgm_key_prompt)
     try:
         controller.authenticate(management_key, touch_callback=prompt_for_touch)
-    except (APDUError, TypeError):
+    except AuthenticationFailed:
+        ctx.fail('Incorrect management key.')
+    except Exception as e:
+        logger.error('Authentication with management key failed.', exc_info=e)
         ctx.fail('Authentication with management key failed.')
 
 
-def _check_algorithm(ctx, controller, algorithm):
-    #  ECCP384 not supported on NEO.
-    if algorithm == ALGO.ECCP384 and controller.version < (4, 0, 0):
-        ctx.fail('ECCP384 is not supported by this YubiKey.')
-    if algorithm == ALGO.RSA1024 and ctx.obj['dev'].is_fips:
-        ctx.fail('RSA1024 is not supported by this YubiKey.')
-
-
-def _check_key_size(ctx, private_key):
-    if ctx.obj['dev'].is_fips and private_key.key_size == 1024:
+def _check_key_size(ctx, controller, private_key):
+    if (private_key.key_size == 1024
+            and ALGO.RSA1024 not in controller.supported_algorithms):
         ctx.fail('1024 is not a supported key size on this YubiKey.')
 
 
-def _check_pin_policy(ctx, controller, pin_policy):
+def _check_pin_policy(ctx, dev, controller, pin_policy):
     if pin_policy is not None and not controller.supports_pin_policies:
         ctx.fail('PIN policy is not supported by this YubiKey.')
-    if ctx.obj['dev'].is_fips and pin_policy == PIN_POLICY.NEVER:
+    if dev.is_fips and pin_policy == PIN_POLICY.NEVER:
         ctx.fail('PIN policy NEVER is not supported by this YubiKey.')
 
 
