@@ -28,12 +28,14 @@
 
 from __future__ import absolute_import
 from enum import IntEnum, unique
-from .driver_ccid import APDUError, SW_OK, SW_APPLICATION_NOT_FOUND
+from .device import YubiKey
+from .driver_ccid import APDUError, SW
 from .util import (
     AID, Tlv, parse_tlvs,
+    is_cve201715361_vulnerable_firmware_version,
     ensure_not_cve201715361_vulnerable_firmware_version)
-from collections import namedtuple
 from cryptography import x509
+from cryptography.exceptions import InvalidSignature
 from cryptography.utils import int_to_bytes, int_from_bytes
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
@@ -90,7 +92,8 @@ class ALGO(IntEnum):
                 return cls.ECCP256
             elif curve_name == 'secp384r1':
                 return cls.ECCP384
-        raise ValueError('Unsupported key type!')
+        raise UnsupportedAlgorithm(
+            'Unsupported key type: %s' % type(key), key=key)
 
     @classmethod
     def from_string(cls, algorithm):
@@ -102,7 +105,9 @@ class ALGO(IntEnum):
             return cls.ECCP256
         if algorithm == 'ECCP384':
             return cls.ECCP384
-        raise ValueError('Unsupported algorithm!')
+        raise UnsupportedAlgorithm(
+            'Unsupported algorithm name: %s' % algorithm,
+            algorithm_id=algorithm)
 
     @classmethod
     def is_rsa(cls, algorithm_int):
@@ -222,7 +227,7 @@ class PIN_POLICY(IntEnum):
             return cls.ONCE
         if pin_policy == 'ALWAYS':
             return cls.ALWAYS
-        raise ValueError('Unsupported pin policy!')
+        raise UnknownPinPolicy(pin_policy)
 
 
 @unique
@@ -242,22 +247,71 @@ class TOUCH_POLICY(IntEnum):
             return cls.ALWAYS
         if touch_policy == 'CACHED':
             return cls.CACHED
-        raise ValueError('Unsupported touch policy!')
+        raise UnknownTouchPolicy(touch_policy)
 
 
-@unique
-class SW(IntEnum):
-    NO_SPACE = 0x6a84
-    COMMAND_ABORTED = 0x6f00
-    MORE_DATA = 0x61
-    INVALID_INSTRUCTION = 0x6d00
-    NOT_FOUND = 0x6a82
-    ACCESS_DENIED = 0x6982
-    AUTHENTICATION_BLOCKED = 0x6983
-    INCORRECT_PARAMETERS = 0x6a80
+class AuthenticationFailed(Exception):
+    def __init__(self, message, sw, applet_version):
+        super().__init__(message)
+        self.tries_left = (
+            tries_left(sw, applet_version)
+            if is_verify_fail(sw, applet_version)
+            else None)
 
 
-CodeChangeResult = namedtuple('CodeChangeResult', ['success', 'tries_left'])
+class AuthenticationBlocked(AuthenticationFailed):
+    def __init__(self, message, sw):
+        # Dummy applet_version since sw will always be "authentication blocked"
+        super().__init__(message, sw, ())
+
+
+class BadFormat(Exception):
+    def __init__(self, message, bad_value):
+        super().__init__(message)
+        self.bad_value = bad_value
+
+
+class KeypairMismatch(Exception):
+    def __init__(self, slot, cert):
+        super().__init__(
+            'The certificate does not match the private key in slot %s.' % slot)
+        self.slot = slot
+        self.cert = cert
+
+
+class UnsupportedAlgorithm(Exception):
+    def __init__(self, message, algorithm_id=None, key=None, ):
+        super().__init__(message)
+        if algorithm_id is None and key is None:
+            raise ValueError(
+                'At least one of algorithm_id and key must be given.')
+
+        self.algorithm_id = algorithm_id
+        self.key = key
+
+
+class UnknownPinPolicy(Exception):
+    def __init__(self, policy_name):
+        super().__init__(
+            'Unsupported pin policy: %s' % policy_name)
+        self.policy_name = policy_name
+
+
+class UnknownTouchPolicy(Exception):
+    def __init__(self, policy_name):
+        super().__init__(
+            'Unsupported touch policy: %s' % policy_name)
+        self.policy_name = policy_name
+
+
+class WrongPin(AuthenticationFailed):
+    def __init__(self, sw, applet_version):
+        super().__init__('Incorrect PIN', sw, applet_version)
+
+
+class WrongPuk(AuthenticationFailed):
+    def __init__(self, sw, applet_version):
+        super().__init__('Incorrect PUK', sw, applet_version)
 
 
 PIN = 0x80
@@ -277,14 +331,18 @@ def _pack_pin(pin):
     if isinstance(pin, six.text_type):
         pin = pin.encode('utf8')
     if len(pin) > 8:
-        raise ValueError('PIN/PUK too large (max 8 bytes, was %d)' % len(pin))
+        raise BadFormat(
+            'PIN/PUK too large (max 8 bytes, was %d)' % len(pin), pin)
     return pin.ljust(8, b'\xff')
 
 
 def _get_key_data(key):
     if isinstance(key, rsa.RSAPrivateKey):
         if key.public_key().public_numbers().e != 65537:
-            raise ValueError('Unsupported RSA exponent!')
+            raise UnsupportedAlgorithm(
+                'Unsupported RSA exponent: %d'
+                % key.public_key().public_numbers().e,
+                key=key)
 
         if key.key_size == 1024:
             algo = ALGO.RSA1024
@@ -293,7 +351,8 @@ def _get_key_data(key):
             algo = ALGO.RSA2048
             ln = 128
         else:
-            raise ValueError('Unsupported RSA key size!')
+            raise UnsupportedAlgorithm(
+                'Unsupported RSA key size: %d' % key.key_size, key=key)
 
         priv = key.private_numbers()
         data = Tlv(0x01, int_to_bytes(priv.p, ln)) + \
@@ -309,11 +368,12 @@ def _get_key_data(key):
             algo = ALGO.ECCP384
             ln = 48
         else:
-            raise ValueError('Unsupported elliptic curve!')
+            raise UnsupportedAlgorithm(
+                    'Unsupported elliptic curve: %s', key.curve, key=key)
         priv = key.private_numbers()
         data = Tlv(0x06, int_to_bytes(priv.private_value, ln))
     else:
-        raise ValueError('Unsupported key type!')
+        raise UnsupportedAlgorithm('Unsupported key type!', key=key)
     return algo, data
 
 
@@ -326,7 +386,8 @@ def _dummy_key(algorithm):
         return ec.generate_private_key(ec.SECP256R1(), default_backend())
     if algorithm == ALGO.ECCP384:
         return ec.generate_private_key(ec.SECP384R1(), default_backend())
-    raise ValueError('Unsupported algorithm!')
+    raise UnsupportedAlgorithm(
+        'Unsupported algorithm: %s' % algorithm, algorithm_id=algorithm)
 
 
 def _pkcs1_15_pad(algorithm, message):
@@ -364,6 +425,27 @@ def generate_random_management_key():
     return os.urandom(24)
 
 
+def is_verify_fail(sw, applet_version):
+    if applet_version < (1, 0, 4):
+        return 0x6300 <= sw <= 0x63ff
+    else:
+        return SW.is_verify_fail(sw)
+
+
+def tries_left(sw, applet_version):
+    if applet_version < (1, 0, 4):
+        if sw == SW.AUTH_METHOD_BLOCKED:
+            return 0
+
+        if not is_verify_fail(sw, applet_version):
+            raise ValueError(
+                'Cannot read remaining tries from status word: %x' % sw)
+
+        return sw & 0xff
+    else:
+        return SW.tries_left(sw)
+
+
 class PivmanData(object):
 
     def __init__(self, raw_data=Tlv(0x80)):
@@ -458,7 +540,7 @@ class PivController(object):
     def puk_blocked(self):
         return self._pivman_data.puk_blocked
 
-    def send_cmd(self, ins, p1=0, p2=0, data=b'', check=SW_OK):
+    def send_cmd(self, ins, p1=0, p2=0, data=b'', check=SW.OK):
         while len(data) > 0xff:
             self._driver.send_apdu(0x10, ins, p1, p2, data[:0xff])
             data = data[0xff:]
@@ -484,7 +566,7 @@ class PivController(object):
             self._pivman_protected_data = PivmanProtectedData(
                 self.get_data(OBJ.PIVMAN_PROTECTED_DATA))
         except APDUError as e:
-            if e.sw == SW_APPLICATION_NOT_FOUND:
+            if e.sw == SW.NOT_FOUND:
                 # No data there, initialise a new object.
                 self._pivman_protected_data = PivmanProtectedData()
             else:
@@ -493,10 +575,14 @@ class PivController(object):
     def verify(self, pin, touch_callback=None):
         try:
             self.send_cmd(INS.VERIFY, 0, PIN, _pack_pin(pin))
-        except APDUError:
-            raise ValueError(
-                'Pin verification failed. {} tries left.'.format(
-                        self.get_pin_tries()))
+        except APDUError as e:
+            if e.sw == SW.AUTH_METHOD_BLOCKED:
+                raise AuthenticationBlocked('PIN is blocked.', e.sw)
+
+            elif is_verify_fail(e.sw, self.version):
+                raise WrongPin(e.sw, self.version)
+
+            raise
 
         if self.has_derived_key and not self._authenticated:
             self.authenticate(
@@ -509,8 +595,18 @@ class PivController(object):
             self.verify(pin, touch_callback)
 
     def change_pin(self, old_pin, new_pin):
-        self.send_cmd(INS.CHANGE_REFERENCE, 0, PIN,
-                      _pack_pin(old_pin) + _pack_pin(new_pin))
+        try:
+            self.send_cmd(INS.CHANGE_REFERENCE, 0, PIN,
+                          _pack_pin(old_pin) + _pack_pin(new_pin))
+        except APDUError as e:
+            if e.sw == SW.AUTH_METHOD_BLOCKED:
+                raise AuthenticationBlocked('PIN is blocked.', e.sw)
+
+            elif is_verify_fail(e.sw, self.version):
+                raise WrongPin(e.sw, self.version)
+
+            raise
+
         if self.has_derived_key:
             if not self._authenticated:
                 self.authenticate(_derive_key(old_pin, self._pivman_data.salt))
@@ -520,22 +616,27 @@ class PivController(object):
         try:
             self.send_cmd(INS.CHANGE_REFERENCE, 0, PUK,
                           _pack_pin(old_puk) + _pack_pin(new_puk))
-            return CodeChangeResult(True, None)
         except APDUError as e:
-            retries = self._parse_tries_left(e.sw)
-            logger.debug('PUK change failed, %d tries remaining', retries,
-                         exc_info=e)
-            return CodeChangeResult(False, retries)
+            if e.sw == SW.AUTH_METHOD_BLOCKED:
+                raise AuthenticationBlocked('PUK is blocked.', e.sw)
+
+            elif is_verify_fail(e.sw, self.version):
+                raise WrongPuk(e.sw, self.version)
+
+            raise
 
     def unblock_pin(self, puk, new_pin):
         try:
             self.send_cmd(
                 INS.RESET_RETRY, 0, PIN, _pack_pin(puk) + _pack_pin(new_pin))
         except APDUError as e:
-            tries = self._parse_tries_left(e.sw)
-            if tries == 0:
-                raise ValueError('PUK is blocked.')
-            raise ValueError('Unblock PIN failed, {} tries left.'.format(tries))
+            if e.sw == SW.AUTH_METHOD_BLOCKED:
+                raise AuthenticationBlocked('PUK is blocked.', e.sw)
+
+            elif is_verify_fail(e.sw, self.version):
+                raise WrongPuk(e.sw, self.version)
+
+            raise
 
     def set_pin_retries(self, pin_retries, puk_retries):
         self.send_cmd(INS.SET_PIN_RETRIES, pin_retries, puk_retries)
@@ -562,7 +663,12 @@ class PivController(object):
         ct1 = self.send_cmd(INS.AUTHENTICATE, ALGO.TDES, SLOT.CARD_MANAGEMENT,
                             Tlv(TAG.DYN_AUTH, Tlv(0x80)))[4:12]
         backend = default_backend()
-        cipher = Cipher(algorithms.TripleDES(key), modes.ECB(), backend)
+        try:
+            cipher_key = algorithms.TripleDES(key)
+        except ValueError:
+            raise BadFormat('Management key must be exactly 24 bytes long, '
+                            'was: {}'.format(len(key)), None)
+        cipher = Cipher(cipher_key, modes.ECB(), backend)
         decryptor = cipher.decryptor()
         pt1 = decryptor.update(ct1) + decryptor.finalize()
         ct2 = os.urandom(8)
@@ -576,6 +682,19 @@ class PivController(object):
                 INS.AUTHENTICATE, ALGO.TDES, SLOT.CARD_MANAGEMENT,
                 Tlv(TAG.DYN_AUTH, Tlv(0x80, pt1) + Tlv(0x81, ct2))
                 )[4:12]
+
+        except APDUError as e:
+            if e.sw == SW.SECURITY_CONDITION_NOT_SATISFIED:
+                raise AuthenticationFailed(
+                    'Incorrect management key', e.sw, self.version)
+
+            logger.error('Failed to authenticate management key.', exc_info=e)
+            raise
+
+        except Exception as e:
+            logger.error('Failed to authenticate management key.', exc_info=e)
+            raise
+
         finally:
             if touch_callback is not None:
                 touch_timer.cancel()
@@ -597,8 +716,10 @@ class PivController(object):
                                  'store_on_device was not True')
 
         if len(new_key) != 24:
-            raise ValueError('Management key must be exactly 24 bytes long, '
-                             'was: {}'.format(len(new_key)))
+            raise BadFormat(
+                'Management key must be exactly 24 bytes long, was: {}'.format(
+                    len(new_key)),
+                new_key)
 
         if store_on_device or (not store_on_device and self.has_stored_key):
             # Ensure we have access to protected data before overwriting key
@@ -638,6 +759,21 @@ class PivController(object):
                     self._pivman_protected_data.get_bytes())
             except APDUError as e:
                 logger.debug("No PIN provided, can't clear key..", exc_info=e)
+        # Update CHUID and CCC if not set
+        try:
+            self.get_data(OBJ.CAPABILITY)
+        except APDUError as e:
+            if e.sw == SW.NOT_FOUND:
+                self.update_ccc()
+            else:
+                logger.debug('Failed to read CCC...', exc_info=e)
+        try:
+            self.get_data(OBJ.CHUID)
+        except APDUError as e:
+            if e.sw == SW.NOT_FOUND:
+                self.update_chuid()
+            else:
+                logger.debug('Failed to read CHUID...', exc_info=e)
 
     def get_pin_tries(self):
         """
@@ -647,26 +783,14 @@ class PivController(object):
         """
         # Verify without PIN gives number of tries left.
         _, sw = self.send_cmd(INS.VERIFY, 0, PIN, check=None)
-        return self._parse_tries_left(sw)
+        return tries_left(sw, self.version)
 
     def _get_puk_tries(self):
         # A failed unblock pin will return number of PUK tries left,
         # but also uses one try.
         _, sw = self.send_cmd(INS.RESET_RETRY, 0, PIN, _pack_pin('')*2,
                               check=None)
-        return self._parse_tries_left(sw)
-
-    def _parse_tries_left(self, sw):
-        # Blocked, 0 tries left.
-        if sw == SW.AUTHENTICATION_BLOCKED:
-            return 0
-        # YK4, NEO with PIV >= 1.0.4
-        if 0x63c0 <= sw <= 0x63cf:
-            return sw & 0xf
-        # PIV applet < 1.04
-        if 0x6300 <= sw & 0x63ff:
-            return sw & 0xff
-        raise ValueError('Failed reading remaining PIN/PUK tries!')
+        return tries_left(sw, self.version)
 
     def _block_pin(self):
         while self.get_pin_tries() > 0:
@@ -701,6 +825,12 @@ class PivController(object):
         if ALGO.is_rsa(algorithm):
             ensure_not_cve201715361_vulnerable_firmware_version(self.version)
 
+        if algorithm not in self.supported_algorithms:
+            raise UnsupportedAlgorithm(
+                'Algorithm not supported on this YubiKey: {}'
+                .format(algorithm),
+                algorithm_id=algorithm)
+
         data = Tlv(TAG.ALGO, six.int2byte(algorithm))
         if pin_policy:
             data += Tlv(TAG.PIN_POLICY, six.int2byte(pin_policy))
@@ -714,13 +844,16 @@ class PivController(object):
                 int_from_bytes(data[0x82], 'big'),
                 int_from_bytes(data[0x81], 'big')
             ).public_key(default_backend())
-        else:
+        elif algorithm in [ALGO.ECCP256, ALGO.ECCP384]:
             curve = ec.SECP256R1 if algorithm == ALGO.ECCP256 else ec.SECP384R1
             return ec.EllipticCurvePublicNumbers.from_encoded_point(
                 curve(),
                 resp[5:]
             ).public_key(default_backend())
-        raise ValueError('Invalid algorithm!')
+
+        raise UnsupportedAlgorithm(
+            'Invalid algorithm: {}'.format(algorithm),
+            algorithm_id=algorithm)
 
     def generate_self_signed_certificate(
             self, slot, public_key, common_name, valid_from, valid_to,
@@ -752,20 +885,7 @@ class PivController(object):
                          exc_info=e)
             raise
 
-        # Verify that the public key used in the certificate
-        # is from the same keypair as the private key.
-        cert_signature = cert.signature
-        cert_bytes = cert.tbs_certificate_bytes
-        if isinstance(public_key, rsa.RSAPublicKey):
-            verifier = public_key.verifier(
-                cert_signature, padding.PKCS1v15(),
-                cert.signature_hash_algorithm)
-        elif isinstance(public_key, ec.EllipticCurvePublicKey):
-            verifier = public_key.verifier(
-                cert_signature, ec.ECDSA(cert.signature_hash_algorithm))
-        verifier.update(cert_bytes)
-        verifier.verify()
-        self.import_certificate(slot, cert)
+        self.import_certificate(slot, cert, verify=False)
 
     def generate_certificate_signing_request(self, slot, public_key, subject,
                                              touch_callback=None):
@@ -792,10 +912,40 @@ class PivController(object):
         self.send_cmd(INS.IMPORT_KEY, algorithm, slot, data)
         return algorithm
 
-    def import_certificate(self, slot, certificate):
+    def import_certificate(self, slot, certificate, verify=False):
         cert_data = certificate.public_bytes(Encoding.DER)
+
+        if verify:
+            # Verify that the public key used in the certificate
+            # is from the same keypair as the private key.
+            try:
+                public_key = certificate.public_key()
+
+                test_data = b'test'
+                test_sig = self.sign(
+                    slot, ALGO.from_public_key(public_key), test_data)
+
+                if isinstance(public_key, rsa.RSAPublicKey):
+                    public_key.verify(
+                        test_sig, test_data, padding.PKCS1v15(),
+                        certificate.signature_hash_algorithm)
+                elif isinstance(public_key, ec.EllipticCurvePublicKey):
+                    public_key.verify(
+                        test_sig, test_data, ec.ECDSA(hashes.SHA256()))
+                else:
+                    raise ValueError('Unknown key type: ' + type(public_key))
+
+            except APDUError as e:
+                if e.sw == SW.INCORRECT_PARAMETERS:
+                    raise KeypairMismatch(slot, certificate)
+                raise
+
+            except InvalidSignature as e:
+                raise KeypairMismatch(slot, certificate)
+
         self.put_data(OBJ.from_slot(slot), Tlv(TAG.CERTIFICATE, cert_data) +
                       Tlv(TAG.CERT_INFO, b'\0') + Tlv(TAG.LRC))
+        self.update_chuid()
 
     def read_certificate(self, slot):
         data = _parse_tlv_dict(self.get_data(OBJ.from_slot(slot)))
@@ -814,7 +964,9 @@ class PivController(object):
 
     def _raw_sign_decrypt(self, slot, algorithm, payload, condition):
         if not condition(len(payload.value)):
-            raise ValueError('Input has invalid length!')
+            raise BadFormat(
+                'Input has invalid length for algorithm %s' % algorithm,
+                len(payload.value))
 
         data = Tlv(TAG.DYN_AUTH, Tlv(0x82) + payload)
         resp = self.send_cmd(INS.AUTHENTICATE, algorithm, slot, data)
@@ -847,12 +999,18 @@ class PivController(object):
         return certs
 
     def update_chuid(self):
+        # Non-Federal Issuer FASC-N
+        # [9999-9999-999999-0-1-0000000000300001]
+        FASC_N=b'\xd4\xe7\x39\xda\x73\x9c\xed\x39\xce\x73\x9d\x83\x68' + \
+               b'\x58\x21\x08\x42\x10\x84\x21\xc8\x42\x10\xc3\xeb'
+        # Expires on: 2030-01-01
+        EXPIRY=b'\x32\x30\x33\x30\x30\x31\x30\x31'
+
         self.put_data(
             OBJ.CHUID,
-            Tlv(0x30, b'\xd4\xe7\x39\xda\x73\x9c\xed\x39\xce\x73\x9d\x83\x68'
-                b'\x58\x21\x08\x42\x10\x84\x21\x38\x42\x10\xc3\xf5') +
+            Tlv(0x30, FASC_N) +
             Tlv(0x34, os.urandom(16)) +
-            Tlv(0x35, b'\x32\x30\x33\x30\x30\x31\x30\x31') +
+            Tlv(0x35, EXPIRY) +
             Tlv(0x3e) +
             Tlv(TAG.LRC)
         )
@@ -939,3 +1097,16 @@ class PivController(object):
                     TOUCH_POLICY.ALWAYS]  # Cached policy was added in 4.3
         else:
             return [policy for policy in TOUCH_POLICY]
+
+    @property
+    def supported_algorithms(self):
+        return [
+            alg for alg in ALGO
+
+            if not alg == ALGO.TDES
+            if not (ALGO.is_rsa(alg) and
+                    is_cve201715361_vulnerable_firmware_version(self.version))
+            if not (alg == ALGO.ECCP384 and self.version < (4, 0, 0))
+            if not (alg == ALGO.RSA1024 and
+                    YubiKey.is_fips_version(self.version))
+        ]
